"""googlesheet tap class."""
from typing import List
from singer_sdk import Tap, Stream
from singer_sdk import typing as th
# JSON schema typing helpers
from tap_cea_googlesheet.client import request, generate_streams_data
from tap_cea_googlesheet.streams import (
    SheetStream, SheetMetaDataStream
)

NoneType = type(None)


class Tapgooglesheet(Tap):
    """googlesheet tap class."""
    name = "tap-googlesheet"

    config_jsonschema = th.PropertiesList(
        th.Property(
            "client_id",
            th.StringType,
            required=True,
            description="The client id generated from credentials"
        ),
        th.Property(
            "client_secret",
            th.StringType,
            required=True,
            description="The client secret from google auth"
        ),
        th.Property(
            "spreadsheet_id",
            th.StringType,
            required=True,
            description="The id of the spreadsheet to read"
        ),
        th.Property(
            "scopes",
            th.ArrayType(
                th.StringType
            ),
            description="Default scopes defined",
            default=[
                'https://www.googleapis.com/auth/drive.readonly',
                'https://www.googleapis.com/auth/drive.metadata.readonly',
                'https://www.googleapis.com/auth/drive.file',
                'https://www.googleapis.com/auth/spreadsheets.readonly']
        ),
        th.Property(
            "sheet_ranges",
            th.ArrayType(
                th.StringType
            ),
            description='''Specific ranges to be read in the format
            SHEETNAME!A1:B2 where A1:B2 is the range choice''',
        ),
        th.Property(
            "auth_uri",
            th.StringType,
            description="google authorization url",
            default="https://accounts.google.com/o/oauth2/auth"
        ),
        th.Property(
            "token_uri",
            th.StringType,
            description="google token generation url",
            default="https://oauth2.googleapis.com/token"
        ),
        th.Property(
            "auth_provider",
            th.StringType,
            default="https://www.googleapis.com/oauth2/v1/certs"
        )

    ).to_dict()
    
    def discover_streams(self) -> List[Stream]:
        """Return a list of discovered streams."""
        streams_data = generate_streams_data(self.config)
        dynamic_streams = [SheetStream(
            tap=self, stream_data=stream_data, request=request) for stream_data in streams_data]
        return [*dynamic_streams, SheetMetaDataStream(tap=self, request=request)]




"""Custom client handling, including googlesheetStream base class."""
from collections import OrderedDict
from typing import List, Union
import os
import singer
from googleapiclient.discovery import build
from google_auth_oauthlib.flow import InstalledAppFlow
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials

NoneType = type(None)
LOGGER = singer.get_logger()


def request(config, stream_name, **kwargs) -> dict:
    """Shows basic usage of the Sheets API.
        Prints values from a sample spreadsheet.
    """
    creds = None
    client_id = config["client_id"]
    client_secret = config['client_secret']
    credentials = {
        "scopes": config["scopes"],
            "spreadsheet_id": config["spreadsheet_id"],
            "installed": {
                "client_id": client_id,
                "project_id": "",
                "auth_uri": config["auth_uri"],
                "token_uri": config["token_uri"],
                "auth_provider_x509_cert_url": config["auth_provider"],
                "client_secret": client_secret}
        }

    if os.path.exists("token.json"):
        creds = Credentials.from_authorized_user_file(
            "token.json", credentials["scopes"])

    # If there are no (valid) credentials available, let the user log in.
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        else:
            flow = InstalledAppFlow.from_client_config(
                credentials, credentials["scopes"])
            creds = flow.run_local_server(port=0,authorization_prompt_message="")
        with open('token.json', 'w') as token:
            token.write(creds.to_json())

    drive_service = build('drive', 'v3', credentials=creds)
    spreadsheet = drive_service.files().copy(fileId=credentials["spreadsheet_id"], body={
        'mimeType': "application/vnd.google-apps.spreadsheet"}).execute()
    service = build('sheets', 'v4', credentials=creds)
    # Call the Sheets API
    sheet = service.spreadsheets()
    if stream_name =="sheets":
        requested = service.spreadsheets().values().batchGet(
                spreadsheetId=spreadsheet["id"],
                ranges=kwargs["ranges"],
                valueRenderOption=kwargs["valueRenderOption"],
                dateTimeRenderOption=kwargs["dateRenderOption"]
            )
    else:
        requested = sheet.get(
            spreadsheetId=spreadsheet["id"],
            ranges=kwargs["ranges"],
            includeGridData=kwargs["includeOption"]
        )
    response = requested.execute()
    return response


def generate_streams_data(config) -> List[dict] :
    '''
    generates the various stream data used to dynamically
    create streams for each sheet
    '''
    stream_data = []
    range_value = config.get("sheet_ranges")
    spreadsheet_metadata = request(
        config, "sheets_metadata", ranges=range_value,
        includeOption=True)
    sheets = spreadsheet_metadata.get('sheets')
    if sheets:
        for sheet in sheets:
            sheet_title = sheet.get('properties', {}).get('title')
            LOGGER.info('sheet_title = {}'.format(sheet_title))
            try:
                result = get_sheet_schema(sheet, sheet_title)
                if result is not None:
                    schema = {"type": "object", "properties": dict(result)}
                    data = {"schema": schema, "name": sheet_title,
                            "primary_keys": list(result.keys())[0] }
                    stream_data.append(data)
            except NotImplementedError  as error:
                LOGGER.warning('{}'.format(error))
                LOGGER.warning('SKIPPING Malformed sheet: {}'.format(sheet_title))
    return stream_data


def get_sheet_schema(sheet, sheet_title) -> Union[dict, NoneType]:
    '''
     Get the schema for each sheet
    '''
    data = next(iter(sheet.get('data', [])), {})
    row_data = data.get('rowData', [])
    # if the row is empty, or if there is either no header or no initial value
    # when length is 1 we cant determine either the column name or the columnType
    if row_data == [] or row_data[0] == {}:
        LOGGER.info(
            "Skipping sheet, unable to generate sheet schema:{}".format(sheet_title))
        return None

    if len(row_data) == 1:
        # if there is only one row, invert the row to column
        row_data.append({"values": row_data[0]["values"][1:]})
        row_data[0]["values"] = [row_data[0]["values"][0]]

    sheet_schema = OrderedDict()
    # Used To determine schema based on column name

    headers = row_data[0].get('values', [])
    # Used To determine type of data
    first_values = row_data[1].get('values', [])
    headers_list = []

    for i, header in enumerate(headers):
        header_val = header.get('formattedValue')
        if header_val:
            try:
                column_name = str(header_val)
                # Having the same name for two different columns affects the schema creation.
                if column_name in headers_list:
                    column_name = '{}_{}'.format(column_name, i)
                headers_list.append(column_name)
                column_property = evaluate_type(
                    first_values[i], sheet_title, column_name)
            except IndexError as error:
                LOGGER.warning('NO VALUE IN 2ND ROW FOR HEADER. SHEET: {}, COL: {}, {}'.format(
                    sheet_title, column_name, error))
                first_values.append({})
        else:
            LOGGER.info('WARNING: SKIPPED COLUMN; NO COLUMN HEADER. SHEET: {}, COL: {} '.format(
                sheet_title, i))
        sheet_schema[column_name] = column_property
    return sheet_schema


def evaluate_type(value, sheet_title, column_name) -> dict:
    '''
        Evaluates the data type for each sheet property and
        returns the types to get_sheet_schema
    '''
    col_property = None
    column_effective_val = value.get('effectiveValue', {})
    if column_effective_val == {}:
        column_effective_type = 'stringValue'
    else:
        column_effective_type = list(column_effective_val.keys())[0]
    if column_effective_type in ("errorType", 'formulaType'):
        raise TypeError("Error in dataType of second row value; SHEET:{}, COL:{}, Type:{}".format(
            sheet_title, column_name, column_effective_type
        ))
    if column_effective_type == "numberValue":
        number_format = value.get("effectiveFormat", {}).get(
            "numberFormat", {}).get('type')
        if number_format in ("DATE_TIME", "DATE", "TIME"):
            col_property = {
                'type': 'string',
                'format': number_format.lower()
            }
        elif number_format == 'TEXT':
            col_property = {'type': 'string'}

        else:
            col_property = {'type': 'number', 'multipleOf': 1e-15}
    else:
        col_property = {'type': 'string'}
    return col_property



"""Stream type classes for tap-googlesheet."""
from pathlib import Path
from typing import Optional, Iterable

import singer
from singer_sdk.streams import Stream


LOGGER = singer.get_logger()

SCHEMAS_DIR = Path(__file__).parent / Path("./schemas")


class SheetStream(Stream):
    """Dynamic stream for creating a sheet stream based on the number of sheets available."""

    def __init__(self, tap, stream_data, request):
        self.stream = stream_data
        self.request = request
        super().__init__(tap)
        self.range_values = self.config.get("sheet_ranges")
        self.primary_keys = self.stream["primary_keys"]
    replication_method = "FULL_TABLE"

    @property
    def schema(self):
        '''
        Get the sheetstream property schema
        '''
        return self.stream["schema"]

    @property
    def name(self):
        '''
        Get the property name
        '''
        return self.stream["name"]

    def get_records(self, context: Optional[dict]) -> Iterable[dict]:
        """Return a generator of row-type dictionary objects.

        The optional `context` argument is used to identify a specific slice of the
        stream if partitioning is required for the stream. Most implementations do not
        require partitioning and should ignore the `context` argument.
        """
        rows = []
        ranges = f'{self.name}!A1:Z100'

        if self.range_values is not None:
            for sheet_range in self.range_values:
                sheet_name = sheet_range.split("!")[0]
                if sheet_name.lower() == self.name.lower():
                    ranges = sheet_range

        get_columns_and_values = self.request(
            self.config, "sheets", ranges=ranges, valueRenderOption="UNFORMATTED_VALUE",
            dateRenderOption="FORMATTED_STRING").get("valueRanges")[0]

        values = get_columns_and_values["values"]

        if len(values) == 1:
            data_keys = [values[0][0]]
            data_values = [values[0][1:]]
        else:
            data_keys = values[0]
            data_values = values[1:]

        for row in data_values:
            value = dict(zip(data_keys, row))
            rows.append(value)
        for row in rows:
            yield row


class SheetMetaDataStream(Stream):
    '''
     Stream for Meta data resulting from calling google sheet api
    '''

    def __init__(self, tap, request) -> None:
        self.request = request
        super().__init__(tap)
    name = "SheetMetadata"
    primary_keys = ["spreadsheetId", "sheetId"]
    replication_method = "FULL_TABLE"
    schema_filepath = SCHEMAS_DIR / "schema_metadata.json"

    def get_records(self, context: Optional[dict]) -> Iterable[dict]:
        """Return a generator of row-type dictionary objects.

        The optional `context` argument is used to identify a specific slice of the
        stream if partitioning is required for the stream. Most implementations do not
        require partitioning and should ignore the `context` argument.
        """
        rows = []
        get_columns_and_values = self.request(
            self.config, "sheets_metadata",ranges=None, includeOption=False)
        for sheet in get_columns_and_values.get('sheets'):
            properties = sheet.get("properties")
            rows.append({
                "spreadsheetId": get_columns_and_values["spreadsheetId"],
                "spreadsheetUrl": get_columns_and_values["spreadsheetUrl"],
                "sheet": {
                    "sheetId": properties["sheetId"],
                    "title": properties["title"],
                    "index": properties["index"],
                    "sheetType": properties["sheetType"]
                }
            })
        for row in rows:
            yield row



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tap_cea_googlesheet', 'tap_cea_googlesheet.tests']

package_data = \
{'': ['*'], 'tap_cea_googlesheet': ['schemas/*']}

install_requires = \
['google-api-python-client==2.31.0',
 'google-auth-oauthlib==0.4.6',
 'google-auth>=2.3.3,<3.0.0',
 'oauth2client>=4.1.3,<5.0.0',
 'pylint>=2.12.2,<3.0.0',
 'requests>=2.25.1,<3.0.0',
 'singer-sdk>=0.3.16,<0.4.0']

entry_points = \
{'console_scripts': ['tap-cea-googlesheet = '
                     'tap_cea_googlesheet.tap:Tapgooglesheet.cli']}

setup_kwargs = {
    'name': 'tap-cea-googlesheet',
    'version': '0.0.4',
    'description': '`tap-cea-googlesheet` is a Singer tap for googlesheet, built with the Meltano SDK for Singer Taps.',
    'long_description': None,
    'author': 'Basil',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<3.10',
}


setup(**setup_kwargs)
